type rectyId = int
type tvid = int
type stype = TVar of tvid
           | UnitT
           | BoolT
           | IntT
           | ChanT of stype
           | PairT of stype * stype
           | FunT of stype * stype
           | RecT of rectyId * stype list 
type tenv = (Syntax.vid, stype) Utilities.env

open Syntax;;

exception TODO;;
exception Fatal of string;;

type constraints = (stype * stype) list
let emptyC = []
let makeEqT t1 t2 = [(t1,t2)]
let merge_constraints = Utilities.list_flatten

let current_tvid = ref 0;;
let new_tvid() = 
    let x = !current_tvid in
    let _ = (current_tvid := !current_tvid+1) in
       x;;
let new_stype() = 
    TVar(new_tvid());;

let tenv_extend x env =
  let tx = new_stype() in
    (Utilities.env_extend x tx env, tx);;

let const_table = [("+", fun () -> FunT(PairT(IntT, IntT),IntT));
                   ("-", fun () -> FunT(PairT(IntT, IntT),IntT));
                   ("*", fun () -> FunT(PairT(IntT, IntT),IntT));
                   ("/", fun () -> FunT(PairT(IntT, IntT),IntT));
                   ("=", fun () -> FunT(PairT(IntT, IntT),BoolT));
                   (">", fun () -> FunT(PairT(IntT, IntT),BoolT));
                   (">=", fun () -> FunT(PairT(IntT, IntT),BoolT));
                   ("<", fun () -> FunT(PairT(IntT, IntT),BoolT));
                   ("<=", fun () -> FunT(PairT(IntT, IntT),BoolT));
                   ("fst", fun () -> let t1 = new_stype() in let t2 = new_stype() in
                                     (FunT(PairT(t1,t2),t1)));
                   ("snd", fun () -> let t1 = new_stype() in let t2 = new_stype() in
                                     (FunT(PairT(t1,t2),t2)));
                   ("nil", fun () -> (RecT(0, [new_stype(); new_stype()])));
                   ("cons", fun () -> let t1 = new_stype() in let t2 = new_stype() in
                              (FunT(PairT(t1, RecT(0, [t2;t2])), RecT(0, [t1;t2]))));
                   ("hd", fun () -> let t1 = new_stype() in let t2 = new_stype() in
                               FunT(RecT(0, [t1;t2]), t1));
                   ("tl", fun () -> let t1 = new_stype() in let t2 = new_stype() in
                               FunT(RecT(0, [t1;t2]), RecT(0, [t2;t2])));
                   ("null", fun () -> let t1 = new_stype() in let t2 = new_stype() in
                               FunT(RecT(0, [t1;t2]), BoolT))];;

let lookup_prim prname = 
  try
    (List.assoc prname const_table)()
  with
    Not_found -> raise (Fatal ("unknown primitive "^prname));;

type constructor_tab = (cid * (unit -> stype)) list
let constructor_table: constructor_tab ref = 
     ref [("nil", fun () -> (RecT(0, [new_stype(); new_stype()])));
      ("cons", fun () -> (RecT(0, [new_stype(); new_stype()])))]

let lookup_constructor con = 
   try 
     (List.assoc con !constructor_table)()
   with
     Not_found -> (raise (Fatal ("unknown constructor "^con)))

let rec tinf p env =
  match p with
    Zero -> (Zero, emptyC)
  | Out(x,v,p1,pc) ->
      let (v', t1,c1) = tinfV v env in
      let (p1',c2) = tinf p1 env in
      let t3 = tinfVar x env in
      let c3 = makeEqT t3 (ChanT(t1)) in
      let c = merge_constraints [c1;c2;c3] in
        (Out(x,v',p1',pc), c)
  | In(x,y,_,p1,pc) ->
      let tx = tinfVar x env in
      let (env1,ty) = tenv_extend y env in
      let (p1',c1) = tinf p1 env1 in
      let c2 = makeEqT tx (ChanT(ty)) in
      let c = merge_constraints [c1;c2] in
        (In(x,y,ty,p1',pc),c)
  | Par(p1,p2) ->
      let (p1',c1) = tinf p1 env in
      let (p2',c2) = tinf p2 env in
      let c = merge_constraints [c1;c2] in
       (Par(p1',p2'), c)
  | Rep(p1) ->
      let (p1',c1) = tinf p1 env in
       (Rep(p1'), c1)
  | Nu(x,_,p1,pc) ->
      let (env1,tx) = tenv_extend x env in
      let (p1',c1) = tinf p1 env1 in
      let t = new_stype() in
      let c2 = makeEqT tx (ChanT t) in
      let c = merge_constraints [c1;c2] in
        (Nu(x,tx,p1',pc), c)
  | If(v, p1, p2, pc) ->
      let (p1',c1) = tinf p1 env in
      let (p2',c2) = tinf p2 env in
      let (v',tv,c3) = tinfV v env in
      let c4 = makeEqT tv BoolT in
      let c = merge_constraints [c1;c2;c3;c4] in
       (If(v',p1',p2',pc), c)
  | Let(x,_,v,p1,pc) ->
      let (v',tx,c1) = tinfV v env in
      let env1 = Utilities.env_extend x tx env in
      let (p1',c2) = tinf p1 env1 in
      let c = merge_constraints [c1;c2] in
        (Let(x,tx,v', p1',pc), c)
  | MatchL(x,matchlist,pc) -> 
      let tx = tinfVar x env in
      let (matchlist',c) = tinfMatchL tx matchlist env in
        (MatchL(x,matchlist',pc), c)
  | Special(p1) ->
      let (p1',c1) = tinf p1 env in
       (Special(p1'), c1)
and tinfMatchL t matchlist env =
   match matchlist with
      [] -> ([], emptyC)
    | (cid, p1)::matchlist1 ->
         let (matchlist1',c1) = tinfMatchL t matchlist1 env in
         let (p1',c2) = tinf p1 env in
         let tc = lookup_constructor cid in
         let c3 = makeEqT t tc in
         let c = merge_constraints [c1;c2;c3] in
            ((cid,p1')::matchlist1', c)
and tinfVar x env =
   try 
     Utilities.env_lookup x env
   with
     Not_found -> raise (Fatal ("unknown variables "^x))
and tinfV v env = 
   match v with
     Var(x) -> (Var(x), tinfVar x env, emptyC)
   | Unit -> (Unit, UnitT, emptyC)
   | SBool(b,sl) -> (SBool(b,sl), BoolT, emptyC)
   | SInt(b,sl) -> (SInt(b,sl), IntT, emptyC)
   | Bool(b) -> (Bool(b), BoolT, emptyC)
   | Int(n) -> (Int n, IntT, emptyC)
   | Prim(s,_) -> let t = lookup_prim s in (Prim(s,Some t), t, emptyC)
   | App(v1,v2, pc) ->
        let (v1',t1, c1) = tinfV v1 env in
        let (v2',t2, c2) = tinfV v2 env in
        let t3 = new_stype() in
        let c3 = makeEqT t1 (FunT(t2, t3)) in
        let c = merge_constraints [c1;c2;c3] in
          (App(v1',v2',pc),t3, c)
   | Pair(v1,v2) -> 
        let (v1',t1, c1) = tinfV v1 env in
        let (v2',t2, c2) = tinfV v2 env in
        let c = merge_constraints [c1;c2] in
          (Pair(v1',v2'),PairT(t1,t2), c)

exception Unif of stype * stype;;
exception Occur_check;;

let rec tsubst s t =
  match t with
    TVar(tv) -> Utilities.subst_var s tv t
  | UnitT -> UnitT
  | BoolT -> BoolT
  | IntT -> IntT
  | ChanT(t1) -> ChanT(tsubst s t1)
  | PairT(t1,t2) -> PairT(tsubst s t1, tsubst s t2)
  | FunT(t1,t2) -> FunT(tsubst s t1, tsubst s t2)
  | RecT(id, tl) -> RecT(id, List.map (tsubst s) tl)

let tsubst_for_proc s = Syntax.map_for_t (tsubst s)

let unif c =
  let rec occur_check tv t =
    match t with
      TVar(tv1) -> if tv=tv1 then raise Occur_check
                   else ()
    | ChanT t1 -> occur_check tv t1
    | PairT(t1,t2) -> (occur_check tv t1; occur_check tv t2)
    | FunT(t1,t2) -> (occur_check tv t1; occur_check tv t2)
    | RecT(id,tl) -> (let _ = List.map (occur_check tv) tl in ())
    | _ -> ()
  in  
  let subst_for_c s c =
     List.map (fun (t1,t2)->(tsubst s t1, tsubst s t2)) c 
  in
  let rec unif_aux (c,s) =
    match c with
       [] -> s
     | (t1,t2)::c1 ->
        if t1=t2 then unif_aux (c1, s)
        else 
         match (t1,t2) with
           (TVar tv1, _) -> 
               ((try occur_check tv1 t2 with Occur_check -> raise (Unif(t1,t2))); 
                let s1 = Utilities.make_subst tv1 t2 in
                let s2 = Utilities.comp_subst tsubst s1 s in
                let c1' = subst_for_c s1 c1 in
                  unif_aux (c1', s2))
         | (_, TVar tv2) -> unif_aux ((t2,t1)::c1, s)
         | (ChanT t1', ChanT t2') -> unif_aux ((t1',t2')::c1, s)
         | (PairT(t11,t12), PairT(t21,t22)) -> 
                 unif_aux ((t11,t21)::(t12,t22)::c1, s)
         | (FunT(t11,t12), FunT(t21,t22)) -> 
                 unif_aux ((t11,t21)::(t12,t22)::c1, s)
         | (RecT(id1,tl1), RecT(id2, tl2)) -> 
              if id1 = id2 then
                 let tpl = List.combine tl1 tl2 in
                 let cl = List.map (fun (t1,t2)->(makeEqT t1 t2)) tpl in
                 let c = merge_constraints (cl@[c1]) in
                    unif_aux (c, s)
              else raise (Unif(t1,t2))
         | _ -> raise (Unif(t1,t2))
  in 
    unif_aux (c, Utilities.subst_empty)

let genC p = tinf p Utilities.env_empty

(*** main function for type inference ***)
(*** elaborate p returns a pair (p', env)
 ***  where p' is the process obtained from p by attaching type annotations to bound variables
 ***        env is the type environment
 ***)
let elaborate p = 
  let fv = Syntax.fv_proc p in   (*** fv <- the set of free variables in p ***)
  let env = List.fold_left 
            (fun env -> fun x -> Utilities.env_extend x (new_stype()) env) 
             Utilities.env_empty
             fv    
     (*** env <- x1:a1,...,xn:an, 
      *** where x1,...,xn are the free variables in p, and 
      ***       a1,...,an are fresh type variables
      *** "env" serves as a template for the type environment of p
      ***)
  in
  let (p', c) = tinf p env in
      (*** p'<-- type-annotated term 
       *** c <-- unification constraints on types
       ***)
  let s = unif c in
      (*** s <-- most general unifier for the unification constraints c ***)
  let p'' = tsubst_for_proc s p' in
      (*** apply the substitution to p' ***)
  let env' = Utilities.env_map (tsubst s) env in
      (*** apply the substitution to env ***)
     (p'', env')

let rectype_print_table =
  ref [(0, fun f -> fun tl -> ((f 3 (List.hd(tl)))^" list"))]
let string_of_type t =
  let rec t2string pr t = 
    match t with
      TVar(tv) -> "'a"^(string_of_int tv)
    | UnitT -> "unit"
    | BoolT -> "bool"
    | IntT -> "int"
    | ChanT(t1) -> "chan("^(t2string 0 t1)^")"
    | PairT(t1,t2) -> 
           (if pr>1 then "(" else "")
           ^(t2string 2 t1)^"*"^(t2string 2 t2)
          ^(if pr>2 then ")" else "")
    | FunT(t1,t2) -> 
           (if pr>0 then "(" else "")
          ^(t2string 1 t1)
          ^"->"
          ^(t2string pr t2)
          ^(if pr>0 then ")" else "")
    | RecT(id, tl) ->
        let f = List.assoc id !rectype_print_table in
          f t2string tl
  in t2string 0 t;;

let print_env env = Utilities.print_env (fun x->x) string_of_type env;;


let rec get_specials p env =
  match p with
    Zero -> []
  | Out(x,v,p1,pc) ->
      get_specials p1 env
  | In(x,y,t,p1,pc) ->
      let env' = Utilities.env_extend y t env in
        get_specials p1 env'
  | Par(p1,p2) ->
      let p1' = get_specials p1 env in
      let p2' = get_specials p2 env in
        p1'@p2'
  | Rep(p1) ->
      get_specials p1 env
  | Nu(x,t,p1,pc) ->
      let env' = Utilities.env_extend x t env in
        get_specials p1 env'
  | If(v, p1, p2, pc) ->
      let p1' = get_specials p1 env in
      let p2' = get_specials p2 env in
        p1'@p2'
  | Let(x,t,v,p1,pc) ->
      let env' = Utilities.env_extend x t env in
        get_specials p1 env'
  | MatchL(x,matchlist,pc) -> 
      let specials = List.map (fun (cid,p1)-> get_specials p1 env) matchlist in
        List.flatten specials
  | Special(p1) ->
      [(p1, env)]
